import logging
from datetime import datetime, timedelta, timezone


from ._ticktick_api import TicktickAPI
from .data.ticktick_payloads import TicktickPayloads
from .data.ticktick_ids import TicktickListIds
from .data.ticktick_list_parameters import TicktickListParameters as tlp
from .task_model import Task
from ._task_utils import _is_task_a_weight_measurement, _is_task_active, dict_to_task, parse_ticktick_tasks

current_date = datetime.now(timezone.utc)
date_two_weeks_ago = (current_date - timedelta(days=14)).strftime("%Y-%m-%d")
date_tomorrow = (current_date + timedelta(days=1)).strftime("%Y-%m-%d")


class TicktickClient:
    """Ticktick client."""
    BASE_URL = TicktickAPI.BASE_URL
    GET_STATE_URL = BASE_URL + "/batch/check/0"
    CRUD_TASK_URL = BASE_URL + "/batch/task"
    MOVE_TASK_URL = BASE_URL + "/batch/taskProject"
    TASK_URL = BASE_URL + "/task"
    HABIT_CHECKINS_URL = BASE_URL + "/habitCheckins/query"
    COMPLETED_TASKS_URL = BASE_URL + f"/project/all/closed?from={date_two_weeks_ago}%2005:00:00&to={date_tomorrow}" \
                                     f"%2004:59:00&status=Completed&limit=500"
    ABANDONED_TASKS_URL = BASE_URL + f"/project/all/closed?from={date_two_weeks_ago}%2005:00:00&to={date_tomorrow}" \
                                     f"%2004:59:00&status=Abandoned&limit=500"
    DELETED_TASKS_URL = BASE_URL + "/project/all/trash/pagination?start=0&limit=500"
    GENERAL_FOCUS_TIME_URL = BASE_URL + "/pomodoros/statistics/heatmap"
    ACTIVE_FOCUS_TIME_URL = BASE_URL + "/pomodoros/statistics/dist"

    def __init__(self,
                 username: str,
                 password: str,
                 ticktick_list_ids: TicktickListIds,
                 api_token: str | None = None,
                 cookies: dict[str, str] | None = None):
        self.ticktick_api = TicktickAPI(username, password, api_token, cookies)
        self.ticktick_data: dict = {}
        self.ticktick_list_ids: TicktickListIds = ticktick_list_ids
        self._cached_raw_active_tasks: list[dict] = []
        self.all_active_tasks: list[Task] = []
        self.active_tasks: list[Task] = []
        self.completed_tasks: list[Task] = []
        self.deleted_tasks: list[Task] = []
        self.abandoned_tasks: list[Task] = []
        self.weight_measurements: list[Task] = []

        self._get_all_tasks()

    def _get_ticktick_data(self):
        """Gets raw data from Ticktick."""
        self.ticktick_data = self.ticktick_api.get(self.GET_STATE_URL).json()

    def _get_all_tasks(self):
        """Gets all tasks from Ticktick."""
        self._get_ticktick_data()

        raw_active_tasks = self.ticktick_data["syncTaskBean"]["update"]
        if raw_active_tasks == self._cached_raw_active_tasks:
            return

        self._cached_raw_active_tasks = raw_active_tasks
        self.all_active_tasks = parse_ticktick_tasks(raw_active_tasks, self.ticktick_list_ids.get_ids())

        self.active_tasks = []
        for task in self.all_active_tasks:
            if _is_task_a_weight_measurement(task, self.ticktick_list_ids):
                self.weight_measurements.append(task)
            elif _is_task_active(task):
                self.active_tasks.append(task)
            else:
                logging.warning(f"Task {task} does not have a valid status")

    def move_task_to_project(self, task: Task, project_id: str):
        """Moves a task from one project (list) to another in Ticktick.

        Args:
            task: Task to move.
            project_id: Project id to move the task to.
        """
        payload = TicktickPayloads.move_task_to_project(task, project_id)
        self.ticktick_api.post(self.MOVE_TASK_URL, data=payload)

    def replace_task_tags(self, task: Task, tags: tuple[str, ...]) -> bool:
        """Replaces the tags of a task in Ticktick.

        Args:
            task: Task to replace the tags.
            tags: Tags to replace the task tags with.

        Returns:
            True if the tags were replaced successfully, False otherwise.
        """
        self._get_all_tasks()
        tasks_raw_data = [rt for rt in self._cached_raw_active_tasks if rt[tlp.ID] == task.ticktick_id]

        if not tasks_raw_data:
            return False

        task_raw_data = tasks_raw_data[0]
        task_raw_data[tlp.TAGS] = tags
        payload = {"update": [task_raw_data]}
        self.ticktick_api.post(self.CRUD_TASK_URL, data=payload)
        return True

    def get_active_tasks(self) -> list[Task]:
        """Gets all active tasks from Ticktick.

        Returns:
            Active tasks.
        """
        self._get_all_tasks()
        return self.active_tasks

    def get_completed_tasks(self) -> list[Task]:
        """Gets all completed tasks from Ticktick.

        Returns:
            Completed tasks.
        """
        logging.info("Getting completed tasks")

        self._get_ticktick_data()
        raw_completed_tasks = self.ticktick_api.get(self.COMPLETED_TASKS_URL).json()
        self.completed_tasks = parse_ticktick_tasks(raw_completed_tasks, self.ticktick_list_ids.get_ids())

        return self.completed_tasks

    def get_deleted_tasks(self) -> list[Task]:
        """Gets all deleted tasks from Ticktick.

        Returns:
            Deleted tasks.
        """
        self._get_ticktick_data()
        raw_deleted_tasks = self.ticktick_api.get(self.DELETED_TASKS_URL).json()["tasks"]
        self.deleted_tasks = parse_ticktick_tasks(raw_deleted_tasks, self.ticktick_list_ids.get_ids())

        return self.deleted_tasks

    def get_abandoned_tasks(self) -> list[Task]:
        """Gets all abandoned tasks from Ticktick.

        Returns:
            Abandoned tasks.
        """
        self._get_ticktick_data()
        raw_abandoned_tasks = self.ticktick_api.get(self.ABANDONED_TASKS_URL).json()
        self.abandoned_tasks = parse_ticktick_tasks(raw_abandoned_tasks, self.ticktick_list_ids.get_ids())

        return self.abandoned_tasks

    def get_task(self, task_id: str) -> Task:
        """Gets task information from Ticktick using the API.

        Args:
            task_id: Ticktick id of the task to get.

        Returns:
            Task or dictionary with the task information.
        """
        task = self.ticktick_api.get(f"{self.TASK_URL}/{task_id}").json()
        return dict_to_task(task)

    def complete_task(self, task: Task):
        """Completes a task in Ticktick using the API."""
        payload = TicktickPayloads.complete_task(task)
        self.ticktick_api.post(self.CRUD_TASK_URL, data=payload)

    def create_task(self, task: Task, column_id: str | None = None) -> str:
        """Creates a task in Ticktick using the API.

        Args:
            task: Task to create.
            column_id: Column id to create the task in. If it is set to None, the task is created in the default column.

        Returns:
            Ticktick id of the created task.
        """
        payload = TicktickPayloads.create_task(task, column_id)
        response = self.ticktick_api.post(self.CRUD_TASK_URL, payload).json()
        return list(response["id2etag"].keys())[0]

    def get_overall_focus_time(self, date: str) -> float:
        """Gets the overall focus time of a day from Ticktick.

        Args:
            date: Date to get the focus time from in the format YYYY-MM-DD.

        Returns:
            General focus time.
        """
        clean_date = date.replace("-", "")
        raw_time = self.ticktick_api.get(f"{self.GENERAL_FOCUS_TIME_URL}/{clean_date}/{clean_date}").json()

        return round(raw_time[0]["duration"] / 60, 2)

    def get_active_focus_time(self, date: str, active_focus_tags: list[str]) -> float:
        """Gets the active focus time of a day from Ticktick.

        Args:
            date: Date to get the focus time from in the format YYYY-MM-DD.

        Returns:
            Active focus time.
        """
        clean_date = date.replace("-", "")
        raw_time = self.ticktick_api.get(f"{self.ACTIVE_FOCUS_TIME_URL}/{clean_date}/{clean_date}").json()
        tag_time = raw_time.get("tagDurations", {})

        active_focus_time = 0
        for tag in active_focus_tags:
            active_focus_time += tag_time.get(tag, 0)

        return round(active_focus_time / 60, 2)

    def get_tasks_by_list(self, list_ids: list[str]) -> list[Task]:
        """Gets all tasks from Ticktick by list ids.

        Args:
            list_ids: List ids to get the tasks from.

        Returns:
            Tasks.
        """
        self._get_all_tasks()
        return [task for task in self.all_active_tasks if task.project_id in list_ids]



from .task_model import Task as Task
from .ticktick_client import TicktickClient as TicktickClient
from ._task_utils import dict_to_task as dict_to_task
from .data.ticktick_ids import TicktickListIds as TicktickListIds



from enum import Enum


class TicktickTaskParameters(Enum):
    ID = "id"
    PROJECT_ID = "projectId"
    SORTORDER = "sortOrder"
    TITLE = "title"
    CONTENT = "content"
    DESC = "desc"
    START_DATE = "startDate"
    DUE_DATE = "dueDate"
    TIMEZONE = "timeZone"
    IS_FLOATING = "isFloating"
    IS_ALL_DAY = "isAllDay"
    REMINDERS = "reminders"
    REPEAT_FIRST_DATE = "repeatFirstDate"
    REPEAT_FLAG = "repeatFlag"
    EX_DATE = "exDate"
    COMPLETED_TIME = "completedTime"
    COMPLETED_USER_ID = "completedUserId"
    REPEAT_TASK_ID = "repeatTaskId"
    PRIORITY = "priority"
    STATUS = "status"
    ITEMS = "items"
    PROGRESS = "progress"
    MODIFIED_TIME = "modifiedTime"
    ETAG = "etag"
    DELETED = "deleted"
    CREATED_TIME = "createdTime"
    CREATOR = "creator"
    TAGS = "tags"
    REPEAT_FROM = "repeatFrom"
    POMODORO_SUMMARIES = "pomodoroSummaries"
    FOCUS_SUMMARIES = "focusSummaries"
    FOCUS_TIME = "focus_time"
    POMO_DURATION = "pomoDuration"
    STOPWATCH_DURATION = "stopwatchDuration"
    COLUMN_ID = "columnId"
    PARENT_ID = "parentId"
    KIND = "kind"



from datetime import datetime

import pytest

from tickthon import Task
from tickthon import TicktickClient


@pytest.fixture(scope="module")
def ticktick_client(ticktick_info):
    return TicktickClient(ticktick_info["username"], ticktick_info["password"], ticktick_info["ticktick_ids"])


def test_get_active_tasks(ticktick_client):
    active_tasks = ticktick_client.get_active_tasks()

    assert len(active_tasks) > 0
    assert len(ticktick_client.all_active_tasks) >= len(active_tasks)


def test_get_deleted_tasks(ticktick_client):
    deleted_tasks = ticktick_client.get_deleted_tasks()
    assert len(deleted_tasks) > 0


def test_get_abandoned_tasks(ticktick_client):
    abandoned_tasks = ticktick_client.get_abandoned_tasks()
    assert len(abandoned_tasks) > 0


def test_get_completed_tasks(ticktick_client):
    completed_tasks = ticktick_client.get_completed_tasks()
    assert len(completed_tasks) > 0


def test_complete_task(ticktick_client):
    task_id = ticktick_client.create_task(Task(title="test-task", created_date="2099-09-09",
                                               ticktick_id="test-id", ticktick_etag="test-etag"))

    task_to_complete = ticktick_client.get_task(task_id)

    ticktick_client.complete_task(task_to_complete)

    completed_task = ticktick_client.get_task(task_to_complete.ticktick_id)
    assert completed_task.status == 2


def test_move_task(ticktick_client):
    inbox_project_id = ticktick_client.ticktick_list_ids.INBOX
    weight_measurements_id = ticktick_client.ticktick_list_ids.WEIGHT_MEASUREMENTS
    
    id_move_task = ticktick_client.create_task(Task(title="Test move task",
                                                    created_date="2099-09-09",
                                                    ticktick_id="test-move-id",
                                                    ticktick_etag="test-move-etag",
                                                    project_id=inbox_project_id))

    move_task = ticktick_client.get_task(id_move_task)
    ticktick_client.move_task_to_project(move_task, weight_measurements_id)

    moved_task = ticktick_client.get_task(id_move_task)
    assert moved_task.project_id == weight_measurements_id

    ticktick_client.complete_task(moved_task)


@pytest.mark.parametrize("task_tags", [tuple(), ("test-existing-tag",)])
def test_replace_task_tags(ticktick_client, task_tags):
    inbox_project_id = ticktick_client.ticktick_list_ids.INBOX
    id_task = ticktick_client.create_task(Task(title="Test replace task tags", created_date="2099-09-09",
                                               ticktick_id="test-tags-id", ticktick_etag="test-tags-etag",
                                               project_id=inbox_project_id, tags=task_tags))

    replace_tags_task = ticktick_client.get_task(id_task)

    ticktick_client.replace_task_tags(replace_tags_task, ("test-tickthon-tag",))
    active_tasks = ticktick_client.get_active_tasks()

    assert [task for task in active_tasks if task.ticktick_id == id_task][0].tags == ("test-tickthon-tag",)

    ticktick_client.complete_task(ticktick_client.get_task(id_task))


def test_get_overall_focus_time(ticktick_client):
    date = datetime.now().strftime("%Y-%m-%d")
    focus_time = ticktick_client.get_overall_focus_time(date)

    assert isinstance(focus_time, float)
    assert focus_time >= 0


def test_get_tasks_by_list(ticktick_client):
    test_list_id = ["687294a88f083d563d177ee6"]
    task_names = ["Test task - c13-bdb6", "Test task - e01f9b37-f6c0", "Test task - cb10f316ee1", "Test Existing Task Static",
                  "Test Existing Task With Missing Data", "Test Existing Task to Update"]
    original_inbox_id = ticktick_client.ticktick_list_ids.INBOX
    ticktick_client.ticktick_list_ids.INBOX = test_list_id[0]
    ticktick_client._cached_raw_active_tasks = []

    tasks = ticktick_client.get_tasks_by_list(test_list_id)

    ticktick_client.ticktick_list_ids.INBOX = original_inbox_id
    assert len(tasks) == len(task_names)
    assert all(task.title in task_names for task in tasks)



from functools import reduce
from dateutil import parser, tz

from tickthon.data.ticktick_ids import TicktickListIds

from .data.ticktick_task_parameters import TicktickTaskParameters as ttp
from .task_model import Task


def parse_ticktick_tasks(raw_tasks: list[dict] | dict, valid_ticktick_lists_ids: list | None = None) \
        -> list[Task]:
    """Parses raw tasks from Ticktick into Task objects.

    Args:
        raw_tasks: Raw tasks from Ticktick.
        valid_ticktick_lists_ids: Ticktick lists ids to filter tasks by. If it is set to None, all tasks are parsed.

    Returns:
        Parsed tasks.
    """
    ticktick_tasks = []
    valid_ticktick_lists = valid_ticktick_lists_ids if valid_ticktick_lists_ids else []

    if not isinstance(raw_tasks, list):
        raw_tasks = [raw_tasks]

    for raw_task in raw_tasks:
        if valid_ticktick_lists and raw_task[ttp.PROJECT_ID.value] not in valid_ticktick_lists:
            continue

        ticktick_tasks.append(dict_to_task(raw_task))
    return ticktick_tasks


def dict_to_task(raw_task: dict) -> Task:
    """Converts a raw task to a Task object.

    Args:
        raw_task: The raw task as dictionary.

    Returns:
        A Task object.
    """
    return Task(ticktick_id=raw_task[ttp.ID.value],
                ticktick_etag=raw_task[ttp.ETAG.value],
                created_date=get_task_date(raw_task.get(ttp.TIMEZONE.value, ""),
                                           raw_task.get(ttp.CREATED_TIME.value, None)),
                status=raw_task[ttp.STATUS.value],
                title=raw_task[ttp.TITLE.value].strip(),
                focus_time=get_focus_time(raw_task),
                deleted=raw_task.get(ttp.DELETED.value, 0),
                tags=tuple(raw_task.get(ttp.TAGS.value, ())),
                project_id=raw_task[ttp.PROJECT_ID.value],
                timezone=raw_task[ttp.TIMEZONE.value],
                due_date=get_task_date(raw_task[ttp.TIMEZONE.value], raw_task.get(ttp.START_DATE.value, None)),
                column_id=raw_task.get(ttp.COLUMN_ID.value, ""),
                parent_id=raw_task.get(ttp.PARENT_ID.value, "")
                )


def get_focus_time(raw_task: dict) -> float:
    """Returns the focus time of a task.

    Args:
        raw_task: The raw task from Ticktick.

    Returns:
        The focus time of the task.
    """
    focus_time = 0.0
    if ttp.FOCUS_SUMMARIES.value in raw_task:
        for focus_summary in raw_task[ttp.FOCUS_SUMMARIES.value]:
            raw_focus_time = reduce(lambda acc, summary: acc + summary[2], focus_summary.get("focuses", []), 0)
            focus_time += raw_focus_time
        focus_time = round(focus_time / 3600, 2)
    elif ttp.FOCUS_TIME.value in raw_task:
        focus_time = float(raw_task[ttp.FOCUS_TIME.value])

    return focus_time


def get_task_date(raw_task_timezone: str, task_date: str | None) -> str:
    """Returns the date of a task taking into account the timezone.

    Args:
        raw_task_timezone: The timezone of the task.
        task_date: The start date of the task.

    Returns:
        Task's date in iso format YYYY-MM-DDTHH:MM:SS+0000., if the task has no start date, returns an empty string.
    """
    if not task_date:
        return ""

    task_timezone = tz.gettz(raw_task_timezone)
    task_raw_date = parser.parse(task_date)

    localized_task_date = task_raw_date.astimezone(task_timezone)
    task_date = localized_task_date.isoformat()

    return task_date


def _is_task_a_weight_measurement(task: Task, ticktick_ids: TicktickListIds) -> bool:
    """Checks if a task is a weight measurement."""
    weight_measurements_list_id = ticktick_ids.WEIGHT_MEASUREMENTS

    if weight_measurements_list_id is None:
        return False

    return task.project_id == weight_measurements_list_id


def _is_task_active(task: Task) -> bool:
    """Checks if a task is active."""
    return task.status == 0 and task.deleted == 0



from datetime import datetime
from typing import Optional, Collection

from ..task_model import Task


class TicktickPayloads:

    @staticmethod
    def get_habits_checkins(habit_list: dict, after_stamp: int) -> dict:
        return {
            "habitIds": list(habit_list.keys()),
            "afterStamp": after_stamp
        }

    @staticmethod
    def _update_task(task_id: str, project_id: str, status: int | None = None, completed_time: str | None = None,
                     tags: Collection[str] | None = None) -> dict:
        payload = {
            "update": [
                {
                    "completedUserId": 114478622,
                    "projectId": project_id,
                    "id": task_id
                }
            ]
        }

        if status:
            payload["update"][0]["status"] = status

        if completed_time:
            payload["update"][0]["completedTime"] = completed_time

        if tags is not None:
            payload["update"][0]["tags"] = tags

        return payload

    @classmethod
    def complete_task(cls, task: Task) -> dict:
        return cls._update_task(task.ticktick_id, task.project_id, status=2, completed_time=f"{datetime.utcnow()}+0000")

    @classmethod
    def update_task_tags(cls, task: Task, tags: Collection[str]) -> dict:
        return cls._update_task(task.ticktick_id, task.project_id, tags=tags)

    @classmethod
    def move_task_to_project(cls, task: Task, project_id: str) -> list[dict]:
        return [
            {
                "fromProjectId": task.project_id,
                "toProjectId": project_id,
                "taskId": task.ticktick_id,
            }
        ]

    @classmethod
    def create_task(cls, task: Task, column_id: Optional[str] = None) -> dict:
        return {
            "add": [
                {
                    "startDate": task.due_date if task.due_date else None,
                    "columnId": column_id,
                    "projectId": task.project_id if task.project_id else None,
                    "title": task.title,
                    "tags": task.tags if task.tags else None,
                    "timeZone": task.timezone if task.timezone else None,
                }
            ],
            "update": [],
            "delete": [],
            "addAttachments": [],
            "updateAttachments": [],
            "deleteAttachments": []
        }



from attr import define
import attr


@define
class TicktickListIds:
    INBOX: str
    TODAY_BACKLOG: str
    WEEK_BACKLOG: str
    MONTH_BACKLOG: str
    WEIGHT_MEASUREMENTS: str

    def get_ids(self) -> list[str]:
        """Returns a list of all backlog values."""
        return [getattr(self, field.name) for field in attr.fields(self.__class__)]



from enum import Enum

from requests import Session, Response


class RequestTypes(Enum):
    """Types of requests that can be sent to the Ticktick API."""
    GET = "GET"
    POST = "POST"


class TicktickAPI:
    """Ticktick API client."""
    BASE_URL = "https://api.ticktick.com/api/v2"
    SIGNIN_URL = BASE_URL + "/user/signon?wc=true&remember=true"

    USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:95.0) Gecko/20100101 Firefox/95.0'
    X_DEVICE_ = '{"platform":"web","os":"Windows 10","device":"Chrome 123.0.0.0","name":"","version":5303,"id":"65f10b61131d8a5bf9e68825","channel":"website","campaign":"","websocket":""}'  # noqa: E501

    SIGNIN_HEADERS = {'User-Agent': USER_AGENT,
                      'x-device': X_DEVICE_,
                      'Content-Type': 'application/json'}

    def __init__(self, username: str,
                 password: str,
                 api_token: str | None = None,
                 cookies: dict[str, str] | None = None):
        self.session = Session()
        self.session.headers.update({"Content-Type": "application/json",
                                     "User-Agent": self.USER_AGENT,
                                     "x-device": self.X_DEVICE_
                                     })

        self.auth_token, self.cookies = self.validate_token(username, password, api_token, cookies)

    def _login(self, user: str, password: str) -> tuple[str, dict[str, str]]:
        """Logs into Ticktick and returns the authentication token.

        Args:
            user: Ticktick username
            password: Ticktick password

        Returns:
            A tuple with the token and the cookie.
        """
        payload = {"username": user, "password": password}
        response = self.session.post(self.SIGNIN_URL, headers=self.SIGNIN_HEADERS, json=payload)
        response.raise_for_status()

        cookies = {name: value for name, value in self.session.cookies.items()}

        return response.json()["token"], cookies

    def validate_token(self,
                       username: str,
                       password: str,
                       api_token: str | None,
                       cookies: dict[str, str] | None) -> tuple[str, dict[str, str]]:
        """Validate the token. If the token is invalid, login again.

        Args:
            username: The username to login with.
            password: The password to login with.
            api_token: The api token to validate.
            cookies: The cookies to use for the request.

        Returns:
            A tuple with the token and refresh token.
        """
        self.session.headers.update({"Authorization": f"Bearer {api_token}"})

        if cookies:
            self.session.cookies.update(cookies)

        current_token_response = self.session.get(self.BASE_URL + "/batch/check/0")

        if current_token_response.ok and cookies and api_token:
            return api_token, cookies

        return self._login(username, password)

    def post(self, url: str, data: dict | list | None = None) -> Response:
        """Sends a POST request to the Ticktick API.

        Args:
            url: URL to send the request to
            data: Data to send in the request. Defaults to None.

        Returns:
            Response from the Ticktick API
        """

        response = self.session.post(url, json=data)
        response.raise_for_status()

        return response

    def get(self, url: str, data: dict | list | None = None) -> Response:
        """Sends a GET request to the Ticktick API.

        Args:
            url: URL to send the request to
            data: Data to send in the request. Defaults to None.

        Returns:
            Response from the Ticktick API
        """

        response = self.session.get(url, json=data)
        response.raise_for_status()

        return response
